# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Read/Write/Interchange methods for `astropy.cosmology`.

**NOT public API**.
"""

# Import to register with the I/O machinery
from . import cosmology, ecsv, html, latex, mapping, model, row, table, yaml
