.. _Numpy broadcasting rules: https://numpy.org/doc/stable/user/basics.broadcasting.html
.. _Generalized World Coordinate System (GWCS): https://gwcs.readthedocs.io/en/latest/
.. _ASDF: https://asdf-standard.readthedocs.io/en/latest/
.. _SIP: https://fits.gsfc.nasa.gov/registry/sip.html
